package ru.dbotthepony.mc.otm.datagen

import net.minecraft.core.HolderSet
import net.minecraft.core.registries.Registries
import net.minecraft.data.worldgen.BootstrapContext
import net.minecraft.resources.ResourceKey
import net.minecraft.tags.BiomeTags
import net.minecraft.tags.BlockTags
import net.minecraft.util.valueproviders.ClampedNormalFloat
import net.minecraft.util.valueproviders.ConstantFloat
import net.minecraft.util.valueproviders.UniformFloat
import net.minecraft.util.valueproviders.UniformInt
import net.minecraft.world.level.levelgen.GenerationStep
import net.minecraft.world.level.levelgen.VerticalAnchor
import net.minecraft.world.level.levelgen.feature.ConfiguredFeature
import net.minecraft.world.level.levelgen.feature.Feature
import net.minecraft.world.level.levelgen.feature.configurations.OreConfiguration
import net.minecraft.world.level.levelgen.feature.configurations.ReplaceBlockConfiguration
import net.minecraft.world.level.levelgen.heightproviders.VeryBiasedToBottomHeight
import net.minecraft.world.level.levelgen.placement.CountPlacement
import net.minecraft.world.level.levelgen.placement.HeightRangePlacement
import net.minecraft.world.level.levelgen.placement.InSquarePlacement
import net.minecraft.world.level.levelgen.placement.PlacedFeature
import net.minecraft.world.level.levelgen.placement.RarityFilter
import net.minecraft.world.level.levelgen.structure.templatesystem.TagMatchTest
import net.neoforged.neoforge.common.world.BiomeModifier
import net.neoforged.neoforge.registries.NeoForgeRegistries
import ru.dbotthepony.mc.otm.core.math.Decimal
import ru.dbotthepony.mc.otm.data.world.BooleanProvider
import ru.dbotthepony.mc.otm.data.world.OneOfFloatProvider
import ru.dbotthepony.mc.otm.worldgen.placement.StandardDeviationHeightProvider
import ru.dbotthepony.mc.otm.registry.game.MBlocks
import ru.dbotthepony.mc.otm.registry.data.MWorldGenFeatures
import ru.dbotthepony.mc.otm.worldgen.feature.BlackHolePlacerFeature
import ru.dbotthepony.mc.otm.worldgen.placement.EnormousPlacement
import ru.dbotthepony.mc.otm.worldgen.placement.EllipsoidPlacement
import ru.dbotthepony.mc.otm.worldgen.placement.WormPlacement

private object ConfiguredFeatures {
	val TRITANIUM_ORE = key("tritanium_ore")
	val TRITANIUM_ORE_SMALL = key("tritanium_ore_small")
	val DILITHIUM = key("dilithium")
	val BLACK_HOLE = key("black_hole")

	private fun key(name: String): ResourceKey<ConfiguredFeature<*, *>> {
		return ResourceKey.create(Registries.CONFIGURED_FEATURE, modLocation(name))
	}
}

fun registerConfiguredFeatures(context: BootstrapContext<ConfiguredFeature<*, *>>) {
	val stone = TagMatchTest(BlockTags.STONE_ORE_REPLACEABLES)
	val deepslate = TagMatchTest(BlockTags.DEEPSLATE_ORE_REPLACEABLES)

	run {
		val target = listOf(
			OreConfiguration.target(stone, MBlocks.TRITANIUM_ORE.defaultBlockState()),
			OreConfiguration.target(deepslate, MBlocks.DEEPSLATE_TRITANIUM_ORE.defaultBlockState()),
		)

		context.register(ConfiguredFeatures.TRITANIUM_ORE, ConfiguredFeature(Feature.ORE, OreConfiguration(target, 9)))
		context.register(ConfiguredFeatures.TRITANIUM_ORE_SMALL, ConfiguredFeature(Feature.REPLACE_SINGLE_BLOCK, ReplaceBlockConfiguration(target)))
		//context.register(ConfiguredFeatures.TRITANIUM_ORE_SMALL, ConfiguredFeature(MWorldGenFeatures.DEBUG_PLACEMENT, DebugPlacerFeature.Config(MBlocks.TRITANIUM_ORE.defaultBlockState())))
	}

	run {
		val target = listOf(
			OreConfiguration.target(stone, MBlocks.DILITHIUM_ORE.defaultBlockState()),
			OreConfiguration.target(deepslate, MBlocks.DEEPSLATE_DILITHIUM_ORE.defaultBlockState()),
		)

		context.register(ConfiguredFeatures.DILITHIUM, ConfiguredFeature(Feature.REPLACE_SINGLE_BLOCK, ReplaceBlockConfiguration(target)))
	}

	context.register(ConfiguredFeatures.BLACK_HOLE, ConfiguredFeature(
		MWorldGenFeatures.BLACK_HOLE_PLACER,
		BlackHolePlacerFeature.Config(Decimal("0.25"), Decimal(1))))
}

private object PlacedFeatures {
	val NORMAL_TRITANIUM = key("normal_tritanium")
	val DEEP_TRITANIUM = key("deep_tritanium")
	val WORM_TRITANIUM = key("worm_tritanium")
	val DILITHIUM = key("dilithium")
	val BLACK_HOLE = key("black_hole")

	private fun key(name: String): ResourceKey<PlacedFeature> {
		return ResourceKey.create(Registries.PLACED_FEATURE, modLocation(name))
	}
}

fun registerPlacedFeatures(context: BootstrapContext<PlacedFeature>) {
	val configured = context.lookup(Registries.CONFIGURED_FEATURE)

	run {
		val ore = configured.getOrThrow(ConfiguredFeatures.TRITANIUM_ORE)

		context.register(PlacedFeatures.NORMAL_TRITANIUM, PlacedFeature(
			ore,
			listOf(
				CountPlacement.of(UniformInt.of(2, 6)),
				InSquarePlacement.spread(),
				HeightRangePlacement.of(StandardDeviationHeightProvider(VerticalAnchor.absolute(10), 15.0))
			)
		))

		context.register(PlacedFeatures.DEEP_TRITANIUM, PlacedFeature(
			ore,
			listOf(
				CountPlacement.of(UniformInt.of(4, 8)),
				InSquarePlacement.spread(),
				HeightRangePlacement.of(VeryBiasedToBottomHeight.of(VerticalAnchor.aboveBottom(4), VerticalAnchor.absolute(0), 16))
			)
		))

		context.register(PlacedFeatures.WORM_TRITANIUM, PlacedFeature(
			configured.getOrThrow(ConfiguredFeatures.TRITANIUM_ORE_SMALL),
			listOf(
				EnormousPlacement(
					chunkScanRange = 24,
					seedMix = 9284343575495L,
					children = listOf(
						RarityFilter.onAverageOnceEvery(300),
						InSquarePlacement.spread(),
						HeightRangePlacement.of(StandardDeviationHeightProvider(VerticalAnchor.absolute(-40), 15.0)),
						WormPlacement(
							length = UniformInt.of(120, 400),
							turnRateXY = WormPlacement.normalDistributedTurnRate(10f),
							turnRateXZ = WormPlacement.normalDistributedTurnRate(60f),
							turnSpeedXZ = WormPlacement.constantTurnRate(10f),
							turnSpeedXY = WormPlacement.constantTurnRate(4f),
							turnChanceXY = BooleanProvider.Unbiased(6),
							turnChanceXZ = BooleanProvider.BiasedLinear(0.6f, 5),
							maxTravelUp = 16,
							maxTravelDown = 16,
						),
						EllipsoidPlacement(
							count = UniformInt.of(15, 30),
							xLength = ConstantFloat.of(14f),
							yLength = ConstantFloat.of(14f),
							zLength = ConstantFloat.of(14f),
							x = ClampedNormalFloat.of(0f, 0.4f, -1f, 1f),
							y = ClampedNormalFloat.of(0f, 0.4f, -1f, 1f),
							z = ClampedNormalFloat.of(0f, 0.4f, -1f, 1f),
						)
					)
				)
			)
		))
	}

	run {
		val ore = configured.getOrThrow(ConfiguredFeatures.DILITHIUM)

		val ringularity = OneOfFloatProvider.of(
			ClampedNormalFloat.of(0.4f, 0.2f, -2f, 2f),
			ClampedNormalFloat.of(-0.4f, 0.2f, -2f, 2f),
		)

		context.register(PlacedFeatures.DILITHIUM, PlacedFeature(
			ore,
			listOf(
				EnormousPlacement(
					chunkScanRange = 6,
					seedMix = 237483209523709234L,
					children = listOf(
						RarityFilter.onAverageOnceEvery(120),
						InSquarePlacement.spread(),
						HeightRangePlacement.of(StandardDeviationHeightProvider(VerticalAnchor.absolute(0), 15.0)),
						EllipsoidPlacement(
							x = ringularity,
							y = ringularity,
							z = ringularity,
							count = UniformInt.of(8000, 16000),
							xLength = UniformFloat.of(30f, 70f),
							yLength = UniformFloat.of(40f, 90f),
							zLength = UniformFloat.of(30f, 70f),
						)
					),
				)
			)
		))
	}

	val blackHole = configured.getOrThrow(ConfiguredFeatures.BLACK_HOLE)

	context.register(PlacedFeatures.BLACK_HOLE, PlacedFeature(
		blackHole,
		listOf(
			RarityFilter.onAverageOnceEvery(1000),
			InSquarePlacement.spread(),
			HeightRangePlacement.uniform(VerticalAnchor.absolute(64), VerticalAnchor.absolute(128))
		)
	))
}

private object BiomeModifiers {
	val ORES = key("ores")
	val BLACK_HOLE = key("black_hole")

	private fun key(name: String): ResourceKey<BiomeModifier> {
		return ResourceKey.create(NeoForgeRegistries.Keys.BIOME_MODIFIERS, modLocation(name))
	}
}

fun registerBiomeModifiers(context: BootstrapContext<BiomeModifier>) {
	val placed = context.lookup(Registries.PLACED_FEATURE)
	val biomes = context.lookup(Registries.BIOME)

	context.register(
		BiomeModifiers.ORES,
		net.neoforged.neoforge.common.world.BiomeModifiers.AddFeaturesBiomeModifier(
			biomes.getOrThrow(BiomeTags.IS_OVERWORLD),
			HolderSet.direct(
				placed.getOrThrow(PlacedFeatures.NORMAL_TRITANIUM),
				placed.getOrThrow(PlacedFeatures.DEEP_TRITANIUM),
				placed.getOrThrow(PlacedFeatures.WORM_TRITANIUM),
				placed.getOrThrow(PlacedFeatures.DILITHIUM),
			),
			GenerationStep.Decoration.UNDERGROUND_ORES
		)
	)

	context.register(
		BiomeModifiers.BLACK_HOLE,
		net.neoforged.neoforge.common.world.BiomeModifiers.AddFeaturesBiomeModifier(
			biomes.getOrThrow(BiomeTags.IS_OVERWORLD),
			HolderSet.direct(
				placed.getOrThrow(PlacedFeatures.BLACK_HOLE)
			),
			GenerationStep.Decoration.SURFACE_STRUCTURES
		)
	)
}
